/*

 ============================================================================
                   The Apache Software License, Version 1.1
 ============================================================================

 Copyright (C) 1999-2002 The Apache Software Foundation. All rights reserved.

 Redistribution and use in source and binary forms, with or without modifica-
 tion, are permitted provided that the following conditions are met:

 1. Redistributions of  source code must  retain the above copyright  notice,
    this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.

 3. The end-user documentation included with the redistribution, if any, must
    include  the following  acknowledgment:  "This product includes  software
    developed  by the  Apache Software Foundation  (http://www.apache.org/)."
    Alternately, this  acknowledgment may  appear in the software itself,  if
    and wherever such third-party acknowledgments normally appear.

 4. The names "Apache Cocoon" and  "Apache Software Foundation" must  not  be
    used to  endorse or promote  products derived from  this software without
    prior written permission. For written permission, please contact
    apache@apache.org.

 5. Products  derived from this software may not  be called "Apache", nor may
    "Apache" appear  in their name,  without prior written permission  of the
    Apache Software Foundation.

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 APACHE SOFTWARE  FOUNDATION  OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY DIRECT,
 INDIRECT, INCIDENTAL, SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL  DAMAGES (INCLU-
 DING, BUT NOT LIMITED TO, PROCUREMENT  OF SUBSTITUTE GOODS OR SERVICES; LOSS
 OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 This software  consists of voluntary contributions made  by many individuals
 on  behalf of the Apache Software  Foundation and was  originally created by
 Stefano Mazzocchi  <stefano@apache.org>. For more  information on the Apache
 Software Foundation, please see <http://www.apache.org/>.

*/
package org.apache.cocoon.components.modules.input;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.thread.ThreadSafe;

import java.util.Map;
import java.util.Iterator;
import java.util.Vector;

/** 
 * Meta module that obtains values from other module and returns all
 * parameters as XML.
 * 
 * <p>Config</p>
 * <pre>
 *   &lt;!-- in cocoon.xconf --&gt;
 *   &lt;input-module name="request" ignore="do-" strip="user."/&gt;
 *
 *   &lt;!-- e.g. in database.xml --&gt;
 *   &lt;mode type="all" name="xmlmeta" ignore="foo." use="foo" strip="f" root="my-root"/&gt;
 * </pre>
 *
 * <p>If present, "ignore" gives a prefix of parameters to ignore,
 * ignore has precedence over the "use" attribute, "strip" a prefix
 * that will be removed from the final parameter names in the produced
 * XML, "use" is a prefix for parameters to include in the XML, and
 * "root" is the name of the root element in the created XML.</p>
 *
 * <p>Input</p>
 * <pre>
 *    foo.one = ['abc']
 *    foo.two = ['def']
 *    foo1 = ['bar']
 *    foo2 = ['one','two','three']
 *    bar = ['rubber duck']
 * </pre>
 *
 * <p>Output</p>
 * <pre> 
 *   &lt;my-root&gt;
 *     &lt;item name="oo1"&gt;bar&lt;/item&gt;
 *     &lt;item name="oo2"&gt;
 *        &lt;value&gt;one&lt;/value&gt;
 *        &lt;value&gt;two&lt;/value&gt;
 *        &lt;value&gt;three&lt;/value&gt;
 *     &lt;/item&gt;
 *   &lt;/my-root&gt;
 * </pre>
 *
 * <p>Should produce Objects that are XMLizable and Document one
 * day. Currently, produces only Strings.</p>
 *
 * @author <a href="mailto:haul@informatik.tu-darmstadt.de">Christian Haul</a>
 * @version CVS $Id: XMLMetaModule.java,v 1.1.4.1 2002/11/17 19:12:49 haul Exp $
 */
public class XMLMetaModule extends AbstractMetaModule {

    protected String rootName = "root";
    protected String ignore = null;
    protected String use = null;
    protected String strip = null;

    final static Vector returnNames;
    static {
        Vector tmp = new Vector();
        tmp.add("XML");
        returnNames = tmp;
    }


    
    public void configure(Configuration config) throws ConfigurationException {

        this.inputConf = config.getChild("input-module");
        this.rootName = this.inputConf.getAttribute("root",this.rootName);
        this.ignore = this.inputConf.getAttribute("ignore",this.ignore);
        this.use = this.inputConf.getAttribute("use",this.use);
        this.strip = this.inputConf.getAttribute("strip",this.strip);
    }




    public Object getAttribute( String name, Configuration modeConf, Map objectModel ) 
        throws ConfigurationException {

        if (!this.initialized.booleanValue()) {
            if (getLogger().isErrorEnabled()) 
                getLogger().error("Uninitialized Component! FAILING");
            return null;
        }
        if (this.defaultInput == null) {
            if (getLogger().isWarnEnabled()) 
                getLogger().warn("No input module given. FAILING");
            return null;
        }

        // obtain correct configuration objects
        // default vs dynamic
        Configuration inputConfig = null;
        String inputName=null;
        String rootName = this.rootName;
        String ignore  = this.ignore;
        String use  = this.use;
        String strip  = this.strip;
        if (modeConf!=null) {
            inputName   = modeConf.getChild("input-module").getAttribute("name",null);
            rootName = modeConf.getAttribute("root",this.rootName);
            ignore  = modeConf.getAttribute("ignore" ,this.ignore );
            use  = modeConf.getAttribute("use" ,this.use );
            strip  = modeConf.getAttribute("strip" ,this.strip );
            if (inputName != null) {
                inputConfig = modeConf.getChild("input-module");
            }
        }

        InputModule input = null;
        if (inputName != null) input = obtainModule(inputName);

        StringBuffer sb = new StringBuffer();
        sb.append('<').append(rootName).append('>');

        Iterator names = getNames(objectModel, 
                                  this.input, this.defaultInput, this.inputConf,
                                  input, inputName, inputConfig);
        while (names.hasNext()){
            String attribute = (String) names.next();
            if ((use == null || attribute.startsWith(use)) &&
                (ignore == null || !attribute.startsWith(ignore))) {
                Object[] values = getValues(attribute, objectModel,
                                            this.input, this.defaultInput, this.inputConf,
                                            input, inputName, inputConfig);

                if (strip != null && attribute.startsWith(strip)) 
                    attribute = attribute.substring(strip.length());
                sb.append("<item name=\"").append(attribute).append("\">");
                if (values.length == 1) {
                    sb.append(values[0]);
                } else {
                    for (int i=0;i<values.length;i++){
                        sb.append("<value>").append(values[i]).append("</value>");
                    }
                }
                sb.append("</item>");
            }
        }
        if (input != null) releaseModule(input);

        sb.append('<').append('/').append(rootName).append('>');

        return sb.toString();        
    }





    public Iterator getAttributeNames( Configuration modeConf, Map objectModel ) 
        throws ConfigurationException {

         if (!this.initialized.booleanValue()) {
            if (getLogger().isErrorEnabled()) 
                getLogger().error("Uninitialized Component! FAILING");
            return null;
        }
        if (this.defaultInput == null) {
            if (getLogger().isWarnEnabled()) 
                getLogger().warn("No input module given. FAILING");
            return null;
        }

        return XMLMetaModule.returnNames.iterator();
   }




    public Object[] getAttributeValues( String name, Configuration modeConf, Map objectModel ) 
        throws ConfigurationException {

        if (!this.initialized.booleanValue()) {
            if (getLogger().isErrorEnabled()) 
                getLogger().error("Uninitialized Component! FAILING");
            return null;
        }
        if (this.defaultInput == null) {
            if (getLogger().isWarnEnabled()) 
                getLogger().warn("No input module given. FAILING");
            return null;
        }
        
        Object[] values = new Object[1];
        values[0] = this.getAttribute(name, modeConf, objectModel);
        return values;
    }

}
