/*

 ============================================================================
                   The Apache Software License, Version 1.1
 ============================================================================

 Copyright (C) 1999-2002 The Apache Software Foundation. All rights reserved.

 Redistribution and use in source and binary forms, with or without modifica-
 tion, are permitted provided that the following conditions are met:

 1. Redistributions of  source code must  retain the above copyright  notice,
    this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.

 3. The end-user documentation included with the redistribution, if any, must
    include  the following  acknowledgment:  "This product includes  software
    developed  by the  Apache Software Foundation  (http://www.apache.org/)."
    Alternately, this  acknowledgment may  appear in the software itself,  if
    and wherever such third-party acknowledgments normally appear.

 4. The names "Apache Cocoon" and  "Apache Software Foundation" must  not  be
    used to  endorse or promote  products derived from  this software without
    prior written permission. For written permission, please contact
    apache@apache.org.

 5. Products  derived from this software may not  be called "Apache", nor may
    "Apache" appear  in their name,  without prior written permission  of the
    Apache Software Foundation.

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 APACHE SOFTWARE  FOUNDATION  OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY DIRECT,
 INDIRECT, INCIDENTAL, SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL  DAMAGES (INCLU-
 DING, BUT NOT LIMITED TO, PROCUREMENT  OF SUBSTITUTE GOODS OR SERVICES; LOSS
 OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 This software  consists of voluntary contributions made  by many individuals
 on  behalf of the Apache Software  Foundation and was  originally created by
 Stefano Mazzocchi  <stefano@apache.org>. For more  information on the Apache
 Software Foundation, please see <http://www.apache.org/>.

*/
package org.apache.cocoon.components.source;

import org.apache.avalon.framework.component.ComponentException;
import org.apache.avalon.framework.component.ComponentManager;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.Processor;
import org.apache.cocoon.caching.PipelineCacheKey;
import org.apache.cocoon.components.CocoonComponentManager;
import org.apache.cocoon.components.EnvironmentStack;
import org.apache.cocoon.components.pipeline.CacheableEventPipeline;
import org.apache.cocoon.components.pipeline.EventPipeline;
import org.apache.cocoon.components.pipeline.StreamPipeline;
import org.apache.cocoon.environment.Environment;
import org.apache.cocoon.environment.ModifiableSource;
import org.apache.cocoon.environment.Source;
import org.apache.cocoon.environment.wrapper.EnvironmentWrapper;
import org.apache.cocoon.util.HashUtil;
import org.apache.cocoon.xml.AbstractXMLConsumer;
import org.apache.cocoon.xml.ContentHandlerWrapper;
import org.apache.cocoon.xml.XMLConsumer;
import org.apache.cocoon.xml.XMLProducer;
import org.apache.log.Logger;
import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.InputSource;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.ext.LexicalHandler;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Map;
import java.net.MalformedURLException;

/**
 * Description of a source which is defined by a pipeline.
 *
 * @author <a href="mailto:cziegeler@apache.org">Carsten Ziegeler</a>
 * @version CVS $Id: SitemapSource.java,v 1.9.2.6 2002/12/04 09:56:25 cziegeler Exp $
 */

public final class SitemapSource
extends AbstractXMLConsumer
implements ModifiableSource {

    /** The last modification date or 0 */
    private long lastModificationDate;

    /** The system id */
    private String systemId;

    /** The uri */
    private String uri;

    /** The current ComponentManager */
    private ComponentManager manager;

    /** The processor */
    private Processor processor;

    /** The pipeline processor */
    private Processor pipelineProcessor;

    /** The environment */
    private EnvironmentWrapper environment;

    /** The prefix for the processing */
    private String prefix;

    /** The <code>EventPipeline</code> */
    private EventPipeline eventPipeline;

    /** The <code>StreamPipeline</code> */
    private StreamPipeline pipeline;

    /** The redirect <code>Source</code> */
    private Source redirectSource;

    /** The <code>SAXException</code> if unable to get resource */
    private ProcessingException exception;

    /** Do I need a refresh ? */
    private boolean needsRefresh;

    /** The unique key for this processing */
    private Object processKey;

    /**
     * Construct a new object
     */
    public SitemapSource(ComponentManager manager,
                         String           uri,
                         Logger           logger)
    throws IOException, ProcessingException {

        Environment env = CocoonComponentManager.getCurrentEnvironment();
        if (env == null) {
            throw new MalformedURLException("The cocoon protocol can not be used outside an environment.");
        }

        this.manager = manager;
        this.setLogger(logger);
        boolean rawMode = false;

        // remove the protocol
        int position = uri.indexOf(':') + 1;
        if (position != 0) {
            // check for subprotocol
            if (uri.startsWith("raw:", position)) {
                position += 4;
                rawMode = true;
            }
        }

        // does the uri point to this sitemap or to the root sitemap?
        if (uri.startsWith("//", position)) {
            position += 2;
            try {
                this.processor = (Processor)this.manager.lookup(Processor.ROLE);
            } catch (ComponentException e) {
                throw new ProcessingException("Cannot get Processor instance", e);
            }
            this.prefix = ""; // start at the root
        } else if (uri.startsWith("/", position)) {
            position ++;
            this.prefix = null;
            // BUG#9288: Get proper processor, passed either by processor
            // or other cocoon source
            this.processor = CocoonComponentManager.getCurrentProcessor();
        } else {
            throw new ProcessingException("Malformed cocoon URI.");
        }

        // create the queryString (if available)
        String queryString = null;
        int queryStringPos = uri.indexOf('?', position);
        if (queryStringPos != -1) {
            queryString = uri.substring(queryStringPos + 1);
            uri = uri.substring(position, queryStringPos);
        } else if (position > 0) {
            uri = uri.substring(position);
        }

        // build the request uri which is relative to the context
        String requestURI = (this.prefix == null ? env.getURIPrefix() + uri : uri);

        // create system ID
        this.systemId = queryString == null ?
            "cocoon://" + requestURI :
            "cocoon://" + requestURI + "?" + queryString;

        this.environment = new EnvironmentWrapper(env, requestURI, queryString, logger, rawMode);
        this.uri = uri;
        this.refresh();
    }

    /**
     * Get the last modification date of the source or 0 if it
     * is not possible to determine the date.
     */
    public long getLastModified() {
        if (this.needsRefresh) {
            this.refresh();
        }
        return this.lastModificationDate;
    }

    /**
     * Get the content length of the source or -1 if it
     * is not possible to determine the length.
     */
    public long getContentLength() {
        return -1;
    }

    /**
     * Return an <code>InputStream</code> object to read from the source.
     */
    public InputStream getInputStream()
      throws ProcessingException, IOException {

        if (this.needsRefresh) {
            this.refresh();
        }
        // VG: Why exception is not thrown in constructor?
        if (this.exception != null) {
            throw this.exception;
        }

        if (this.redirectSource != null) {
            return this.redirectSource.getInputStream();
        }

        try {
            CocoonComponentManager.enterEnvironment(this.environment,
                                                    this.environment.getObjectModel(),
                                                    this.pipelineProcessor);
            ByteArrayOutputStream os = new ByteArrayOutputStream();
            this.environment.setOutputStream(os);
            this.pipeline.process(this.environment);
            return new ByteArrayInputStream(os.toByteArray());
        } catch (ProcessingException e) {
            throw e;
        } catch (Exception e) {
            throw new ProcessingException("Exception during processing of " + this.systemId, e);
        } finally {
            CocoonComponentManager.leaveEnvironment();
           // Unhide wrapped environment output stream
            this.environment.setOutputStream(null);
            reset();
        }
    }

    /**
     * Return the unique identifer for this source
     */
    public String getSystemId() {
        return this.systemId;
    }

    /**
     * Refresh this object and update the last modified date
     * and content length.
     */
    public void refresh() {
        reset();
        try {
            // initialize the pipelines
            this.processKey = CocoonComponentManager.startProcessing(this.environment);
            this.eventPipeline = (EventPipeline)this.manager.lookup(EventPipeline.ROLE);
            this.pipeline = (StreamPipeline)this.manager.lookup(StreamPipeline.ROLE);

            this.pipeline.setEventPipeline(eventPipeline);
            // set dummy consumer
            ((XMLProducer)eventPipeline).setConsumer(this);

            this.environment.setURI(this.prefix, this.uri);
            this.processor.process(this.environment, pipeline, eventPipeline);
            // BUG#9288: Get processor which was used to build this pipeline
            this.pipelineProcessor = this.environment.changeToLastContext();
            String redirectURL = this.environment.getRedirectURL();
            try {
                // BUG#9288: Set proper processor before setting up the pipeline
                // (will be needed if pipeline has CocoonSource, see <init>)
                CocoonComponentManager.enterEnvironment(this.environment,
                                                        this.environment.getObjectModel(),
                                                        this.pipelineProcessor);
                if (redirectURL == null) {
                    if (this.eventPipeline.getGenerator() != null &&
                        this.eventPipeline instanceof CacheableEventPipeline)
                    {
                        CacheableEventPipeline cep = (CacheableEventPipeline)this.eventPipeline;
                        PipelineCacheKey pck = cep.generateKey(this.environment);
                        Map validity = null;
                        if (pck != null) {
                            validity = cep.generateValidity(this.environment);
                            if (validity != null) {
                                // the event pipeline is cacheable
                                // now calculate a last modification date
                                String hashKey = pck.toString() + validity.toString();
                                this.lastModificationDate = HashUtil.hash(hashKey);
                            }
                        }
                    }
               } else {
                    if (redirectURL.indexOf(":") == -1) {
                        redirectURL = "cocoon:/" + redirectURL;
                    }
                    this.redirectSource = this.environment.resolve(redirectURL);
                    this.lastModificationDate = this.redirectSource.getLastModified();
                }
            } finally {
                CocoonComponentManager.leaveEnvironment();
            }
        } catch (ProcessingException e) {
            reset();
            this.exception = e;
        } catch (Exception e) {
            reset();
            this.exception = new ProcessingException("Could not get sitemap source "
                                                     + this.systemId, e);
        }
        this.needsRefresh = false;
    }

    /**
     * Return a new <code>InputSource</code> object
     */
    public InputSource getInputSource()
    throws ProcessingException, IOException {
        InputSource newObject = new InputSource(this.getInputStream());
        newObject.setSystemId(this.systemId);
        return newObject;
    }

    /**
     * Stream content to the content handler
     */
    public void toSAX(ContentHandler contentHandler)
        throws SAXException, ProcessingException
    {
        if (this.needsRefresh) {
            this.refresh();
        }
        if (this.exception != null) {
            throw this.exception;
        }
        try {
            if (this.redirectSource != null) {
                this.redirectSource.toSAX(contentHandler);
            } else {
                XMLConsumer consumer;
                if (contentHandler instanceof XMLConsumer) {
                    consumer = (XMLConsumer)contentHandler;
                } else if (contentHandler instanceof LexicalHandler) {
                    consumer = new ContentHandlerWrapper(contentHandler, (LexicalHandler)contentHandler);
                } else {
                    consumer = new ContentHandlerWrapper(contentHandler);
                }
                // We have to add an environment changer
                // clean environment stack handling.
                EnvironmentStack envStack = CocoonComponentManager.getCurrentEnvironmentStack();
                int currentOffset = envStack.getOffset();
                try {
                    CocoonComponentManager.enterEnvironment(this.environment,
                                                            this.environment.getObjectModel(),
                                                            this.pipelineProcessor);
                    ((XMLProducer)eventPipeline).setConsumer(new EnvironmentChanger(consumer, envStack));
                    eventPipeline.process(this.environment);
                } finally {
                    CocoonComponentManager.leaveEnvironment();
                    envStack.resetOffset(currentOffset);
                }
            }
        } catch (ComponentException cme) {
            throw new ProcessingException("Could not lookup pipeline components", cme);
        } catch (ProcessingException e) {
            // Preserve original exception
            throw e;
        } catch (Exception e) {
            throw new ProcessingException("Exception during processing of "
                                          + this.systemId, e);
        } finally {
            reset();
        }
    }

    private void reset() {
        if (this.eventPipeline != null) this.manager.release(this.eventPipeline);
        if (this.pipeline != null) this.manager.release(this.pipeline);
        if (this.processKey != null) {
            CocoonComponentManager.endProcessing(this.environment, this.processKey);
            this.processKey = null;
        }
        this.eventPipeline = null;
        this.pipeline = null;
        this.lastModificationDate = 0;
        this.environment.reset();
        if (this.redirectSource != null) this.redirectSource.recycle();
        this.redirectSource = null;
        this.exception = null;
        this.needsRefresh = true;
    }

    public void recycle() {
        reset();
    }
}


/**
 * This class is an {@link XMLConsumer} that changes the current environment.
 * When a pipeline calls an internal pipeline, two environments are
 * established: one for the calling pipeline and one for the internal pipeline.
 * Now, if SAX events are send from the internal pipeline, they are
 * received by some component of the calling pipeline, so inbetween we
 * have to change the environment forth and back.
 */
final class EnvironmentChanger
implements XMLConsumer {

    final XMLConsumer consumer;
    final EnvironmentStack stack;
    
    EnvironmentChanger(XMLConsumer consumer, EnvironmentStack es) {
        this.consumer = consumer;
        this.stack = es;
    }
    
    public void setDocumentLocator(Locator locator) {
        this.stack.incOffset();
        this.consumer.setDocumentLocator(locator);
        this.stack.decOffset();
    }

    public void startDocument()
    throws SAXException {
        this.stack.incOffset();
        this.consumer.startDocument();
        this.stack.decOffset();
    }

    public void endDocument()
    throws SAXException {
        this.stack.incOffset();
        this.consumer.endDocument();
        this.stack.decOffset();
    }

    public void startPrefixMapping(String prefix, String uri)
    throws SAXException {
        this.stack.incOffset();
        this.consumer.startPrefixMapping(prefix, uri);
        this.stack.decOffset();
    }

    public void endPrefixMapping(String prefix)
    throws SAXException {
        this.stack.incOffset();
        this.consumer.endPrefixMapping(prefix);
        this.stack.decOffset();
    }

    public void startElement(String uri, String loc, String raw, Attributes a)
    throws SAXException {
        this.stack.incOffset();
        this.consumer.startElement(uri, loc, raw, a);
        this.stack.decOffset();
    }


    public void endElement(String uri, String loc, String raw)
    throws SAXException {
        this.stack.incOffset();
        this.consumer.endElement(uri, loc, raw);
        this.stack.decOffset();
    }
    
    public void characters(char c[], int start, int len)
    throws SAXException {
        this.stack.incOffset();
        this.consumer.characters(c, start, len);
        this.stack.decOffset();
    }

    public void ignorableWhitespace(char c[], int start, int len)
    throws SAXException {
        this.stack.incOffset();
        this.consumer.ignorableWhitespace(c, start, len);
        this.stack.decOffset();
    }

    public void processingInstruction(String target, String data)
    throws SAXException {
        this.stack.incOffset();
        this.consumer.processingInstruction(target, data);
        this.stack.decOffset();
    }

    public void skippedEntity(String name)
    throws SAXException {
        this.stack.incOffset();
        this.consumer.skippedEntity(name);
        this.stack.decOffset();
    }

    public void startDTD(String name, String publicId, String systemId)
    throws SAXException {
        this.stack.incOffset();
        this.consumer.startDTD(name, publicId, systemId);
        this.stack.decOffset();
    }

    public void endDTD()
    throws SAXException {
        this.stack.incOffset();
        this.consumer.endDTD();
        this.stack.decOffset();
    }

    public void startEntity(String name)
    throws SAXException {
        this.stack.incOffset();
        this.consumer.startEntity(name);
        this.stack.decOffset();
    }

    public void endEntity(String name)
    throws SAXException {
        this.stack.incOffset();
        this.consumer.endEntity(name);
        this.stack.decOffset();
    }

    public void startCDATA()
    throws SAXException {
        this.stack.incOffset();
        this.consumer.startCDATA();
        this.stack.decOffset();
    }

    public void endCDATA()
    throws SAXException {
        this.stack.incOffset();
        this.consumer.endCDATA();
        this.stack.decOffset();
    }

    public void comment(char ch[], int start, int len)
    throws SAXException {
        this.stack.incOffset();
        this.consumer.comment(ch, start, len);
        this.stack.decOffset();
    }
}

