/*

 ============================================================================
                   The Apache Software License, Version 1.1
 ============================================================================

 Copyright (C) 1999-2002 The Apache Software Foundation. All rights reserved.

 Redistribution and use in source and binary forms, with or without modifica-
 tion, are permitted provided that the following conditions are met:

 1. Redistributions of  source code must  retain the above copyright  notice,
    this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.

 3. The end-user documentation included with the redistribution, if any, must
    include  the following  acknowledgment:  "This product includes  software
    developed  by the  Apache Software Foundation  (http://www.apache.org/)."
    Alternately, this  acknowledgment may  appear in the software itself,  if
    and wherever such third-party acknowledgments normally appear.

 4. The names "Apache Cocoon" and  "Apache Software Foundation" must  not  be
    used to  endorse or promote  products derived from  this software without
    prior written permission. For written permission, please contact
    apache@apache.org.

 5. Products  derived from this software may not  be called "Apache", nor may
    "Apache" appear  in their name,  without prior written permission  of the
    Apache Software Foundation.

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 APACHE SOFTWARE  FOUNDATION  OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY DIRECT,
 INDIRECT, INCIDENTAL, SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL  DAMAGES (INCLU-
 DING, BUT NOT LIMITED TO, PROCUREMENT  OF SUBSTITUTE GOODS OR SERVICES; LOSS
 OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 This software  consists of voluntary contributions made  by many individuals
 on  behalf of the Apache Software  Foundation and was  originally created by
 Stefano Mazzocchi  <stefano@apache.org>. For more  information on the Apache
 Software Foundation, please see <http://www.apache.org/>.

*/
package org.apache.cocoon.sitemap;

import org.apache.avalon.framework.activity.Disposable;
import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.component.ComponentManager;
import org.apache.avalon.framework.component.Composable;
import org.apache.avalon.framework.configuration.Configurable;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.logger.AbstractLoggable;
import org.apache.avalon.framework.thread.ThreadSafe;

import org.apache.avalon.excalibur.component.RoleManager;
import org.apache.avalon.excalibur.logger.LogKitManageable;
import org.apache.avalon.excalibur.logger.LogKitManager;

import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.components.pipeline.EventPipeline;
import org.apache.cocoon.components.pipeline.StreamPipeline;
import org.apache.cocoon.environment.Environment;

import java.util.HashMap;
import java.util.Iterator;

/**
 * This class manages all sub <code>Sitemap</code>s of a <code>Sitemap</code>
 * Invokation of sub sitemaps will be done by this instance as well
 * checking regeneration of the sub <code>Sitemap</code>
 *
 * @author <a href="mailto:Giacomo.Pati@pwr.ch">Giacomo Pati</a>
 * @version CVS $Id: Manager.java,v 1.9.2.1 2002/09/21 03:01:13 vgritsenko Exp $
 */
public class Manager extends AbstractLoggable
        implements Component, Configurable, Composable, Contextualizable,
                   ThreadSafe, LogKitManageable, Disposable
{
    protected Context context;

    /** The configuration */
    protected Configuration conf;

    /** The component manager */
    protected ComponentManager manager;

    /** The sitemap role manager */
    protected RoleManager sitemapRoles;

    /** The sitemap logkit manager */
    protected LogKitManager sitemapLogKitManager;

    protected Class handlerClass = Handler.class;

    /**
     * Set the role manager
     */
    public void setRoleManager(RoleManager roles) {
        this.sitemapRoles = roles;
    }

    /**
     * Set the logkit manager
     */
    public void setLogKitManager(LogKitManager logkit) {
        this.sitemapLogKitManager = logkit;
    }

    /**
     * Get a configuration
     * @param conf the configuration
     */
    public void configure(Configuration conf)
    throws ConfigurationException {
        this.conf = conf;

        String value = conf.getAttribute("handler-class", null);
        if (value != null) {
            try {
                handlerClass = Class.forName(value);
            } catch (ClassNotFoundException ex) {
                throw new ConfigurationException("Cannot find Handler class "
                        + value);
            }
        }
    }

    /** get a context
     * @param context the context object
     */
    public void contextualize(Context context) throws ContextException {
        this.context = context;
    }

    /**
     * Get a component manager
     * @param manager the component manager
     */
    public void compose(ComponentManager manager) {
        this.manager = manager;
    }

    /**
     * Invokes the sitemap handler to process a request
     * @param environment the environment
     * @param uri_prefix the prefix to the URI
     * @param source the source of the sitemap
     * @throws Exception there may be several excpetions thrown
     * @return states if the requested resource was produced
     */
    public boolean invoke(Handler sitemapHandler,
                          Environment environment,
                          String uri_prefix,
                          String source)
    throws Exception {
        // Setup to invoke the processing
        setupProcessing(sitemapHandler, environment);
        String prefix = environment.getURIPrefix();
        String uri    = environment.getURI();
        try {
            environment.changeContext(uri_prefix, source);
            return sitemapHandler.process(environment);
        } finally{
            environment.setContext(prefix, uri);
        }
    }

    /**
     * Invokes the sitemap handler to process a request
     * @param environment the environment
     * @param uri_prefix the prefix to the URI
     * @param source the source of the sitemap
     * @throws Exception there may be several excpetions thrown
     * @return states if the requested resource was produced
     */
    public boolean invoke(Handler sitemapHandler,
                          Environment environment,
                          String uri_prefix,
                          String source,
                          StreamPipeline pipeline,
                          EventPipeline eventPipeline)
    throws Exception {
        // Setup to invoke the processing
        setupProcessing(sitemapHandler, environment);
        String prefix = environment.getURIPrefix();
        String uri    = environment.getURI();
        try {
            environment.changeContext(uri_prefix, source);
            return sitemapHandler.process(environment, pipeline, eventPipeline);
        } finally{
            environment.setContext(prefix, uri);
        }
    }

    private void setupProcessing(Handler sitemapHandler, Environment environment)
            throws Exception
    {
        if (!sitemapHandler.available() || sitemapHandler.hasChanged()) {
            sitemapHandler.regenerate(environment);
        }

        if (!sitemapHandler.available()) {
            if (sitemapHandler.getException() != null) {
                throw sitemapHandler.getException();
            } else {
                throw new ProcessingException("The sitemap handler's sitemap is not available. " +
                        "Please check logs for the exact error.");
            }
        }
    }

    protected Handler createHandler(final ComponentManager newManager,
                                 final String source,
                                 final boolean check_reload,
                                 final boolean reload_async,
                                 final long check_delay)
    throws Exception {
        Handler sitemapHandler = null;
        synchronized (this) {
            sitemapHandler = (Handler)handlerClass.newInstance();
            sitemapHandler.setSourceFileName(source);
            sitemapHandler.setCheckReload(check_reload);
            sitemapHandler.setCheckDelay(check_delay);
            sitemapHandler.setLogger(getLogger());
            sitemapHandler.compose(newManager);
            sitemapHandler.contextualize(this.context);
            if (sitemapHandler instanceof Configurable)
                ((Configurable)sitemapHandler).configure(conf);
        }
        return sitemapHandler;
    }

    /**
     * Disposable
     */
    public void dispose() {
        context = null;
        conf = null;
        manager = null;
    }
}
