/*

 ============================================================================
                   The Apache Software License, Version 1.1
 ============================================================================

 Copyright (C) 1999-2002 The Apache Software Foundation. All rights reserved.

 Redistribution and use in source and binary forms, with or without modifica-
 tion, are permitted provided that the following conditions are met:

 1. Redistributions of  source code must  retain the above copyright  notice,
    this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.

 3. The end-user documentation included with the redistribution, if any, must
    include  the following  acknowledgment:  "This product includes  software
    developed  by the  Apache Software Foundation  (http://www.apache.org/)."
    Alternately, this  acknowledgment may  appear in the software itself,  if
    and wherever such third-party acknowledgments normally appear.

 4. The names "Apache Cocoon" and  "Apache Software Foundation" must  not  be
    used to  endorse or promote  products derived from  this software without
    prior written permission. For written permission, please contact
    apache@apache.org.

 5. Products  derived from this software may not  be called "Apache", nor may
    "Apache" appear  in their name,  without prior written permission  of the
    Apache Software Foundation.

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 APACHE SOFTWARE  FOUNDATION  OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY DIRECT,
 INDIRECT, INCIDENTAL, SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL  DAMAGES (INCLU-
 DING, BUT NOT LIMITED TO, PROCUREMENT  OF SUBSTITUTE GOODS OR SERVICES; LOSS
 OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 This software  consists of voluntary contributions made  by many individuals
 on  behalf of the Apache Software  Foundation and was  originally created by
 Stefano Mazzocchi  <stefano@apache.org>. For more  information on the Apache
 Software Foundation, please see <http://www.apache.org/>.

*/
package org.apache.cocoon.sitemap;

import org.apache.avalon.excalibur.component.DefaultRoleManager;
import org.apache.avalon.excalibur.component.RoleManager;
import org.apache.avalon.excalibur.logger.LogKitManager;
import org.apache.avalon.excalibur.xml.Parser;

import org.apache.avalon.framework.component.ComponentManager;
import org.apache.avalon.framework.configuration.Configurable;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.configuration.SAXConfigurationHandler;

import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.Processor;
import org.apache.cocoon.components.language.generator.CompiledComponent;
import org.apache.cocoon.components.pipeline.EventPipeline;
import org.apache.cocoon.components.pipeline.StreamPipeline;
import org.apache.cocoon.components.source.SourceHandler;
import org.apache.cocoon.environment.Environment;
import org.apache.cocoon.util.ClassUtils;

import org.xml.sax.InputSource;

import java.io.InputStream;
import java.util.HashMap;

/**
 * A <code>Processor</code> based on the sitemap language files.
 *
 * @author <a href="mailto:sylvain@apache.org">Sylvain Wallez</a>
 * @author <a href="mailto:vgritsenko@apache.org">Vadim Gritsenko</a>
 * @version CVS $Id: SitemapManager.java,v 1.10.2.1 2002/09/21 03:01:13 vgritsenko Exp $
 */
public class SitemapManager extends Manager implements Processor, Configurable {

    /** The sitemap file */
    protected String sitemapFileName;

    /** The root sitemap's handler */
    protected Handler sitemapHandler;

    public void configure(Configuration sconf) throws ConfigurationException {

        super.configure(sconf);

        this.sitemapFileName = sconf.getAttribute("file");
        boolean checkReload = sconf.getAttributeAsBoolean("check-reload", true);
        boolean reloadAsynchron = sconf.getAttributeAsBoolean("reload-method", false);
        long checkDelay = sconf.getAttributeAsLong("check-delay", 1L) * 1000;

        if (this.getLogger().isDebugEnabled()) {
            getLogger().debug("Sitemap location = " + this.sitemapFileName);
            getLogger().debug("Checking sitemap reload = " + checkReload);
            getLogger().debug("Reloading sitemap asynchron = " + reloadAsynchron);
            getLogger().debug("Sitemap check delay = " + checkDelay + " sec");
        }

        // Read sitemap roles
        Parser p = null;
        Configuration rolesConfig;
        try {
            p = (Parser)this.manager.lookup(Parser.ROLE);
            SAXConfigurationHandler b = new SAXConfigurationHandler();
            InputStream inputStream = ClassUtils.getResource("org/apache/cocoon/sitemap/sitemap.roles").openStream();
            InputSource is = new InputSource(inputStream);
            is.setSystemId("org/apache/cocoon/sitemap/sitemap.roles");
            p.parse(is, b);
            rolesConfig = b.getConfiguration();
        } catch (Exception e) {
            getLogger().error("Could not configure Cocoon environment", e);
            throw new ConfigurationException("Error trying to load configurations", e);
        } finally {
            this.manager.release(p);
        }

        DefaultRoleManager sitemapRoleManager = new DefaultRoleManager();
        sitemapRoleManager.setLogger(getLogger());
        sitemapRoleManager.configure(rolesConfig);

        AbstractSitemap.setRoleManager(sitemapRoleManager, sconf);
        AbstractSitemap.setLogKitManager(this.sitemapLogKitManager);

        try {
            sitemapHandler = createHandler(this.manager, this.sitemapFileName, checkReload, reloadAsynchron, checkDelay);
        } catch (Exception e) {
            throw new ConfigurationException("Failed to create sitemap handler", e);
        }
    }

    /**
     * Process the given <code>Environment</code> to generate the sitemap.
     */
    public void generateSitemap(Environment environment) throws Exception {
        this.sitemapHandler.regenerate(environment);
    }

    /**
     * Process the given <code>Environment</code> producing the output.
     */
    public boolean process(Environment environment) throws Exception {
        return invoke(sitemapHandler, environment, "", sitemapFileName);
    }

    /**
     * Process the given <code>Environment</code> to assemble
     * a <code>StreamPipeline</code> and an <code>EventPipeline</code>.
     */
    public boolean process(Environment environment,
                           StreamPipeline pipeline,
                           EventPipeline eventPipeline)
    throws Exception {
        return invoke(sitemapHandler, environment, "", sitemapFileName, pipeline, eventPipeline);
    }

    public void dispose() {
        this.sitemapHandler.dispose();
        this.sitemapHandler = null;
        super.dispose();
    }
}
